# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import os

from .base import BaseColorer


def get_term_colorer(*a, **kw):
    if os.environ.get('ANSI_COLORS_DISABLED') is not None:
        return None
    term = os.environ.get('TERM')
    if term and '256color' in term:
        return Xterm256Colorer(*a, **kw)
    return Xterm16Colorer(*a, **kw)


class Xterm16Colorer(BaseColorer):
    _colors = dict(
        grey='0', red='1', green='2', yellow='3',
        blue='4', magenta='5', cyan='6', white='7',
        hi_grey='0', hi_red='1', hi_green='2', hi_yellow='3',
        hi_blue='4', hi_magenta='5', hi_cyan='6', hi_white='7')
    _reset = '0'
    _attrs = {
        'bold': '1', 'dark': '2', 'underline': '4', 'blink': '5',
        'reverse': '7', 'concealed': '8',
    }

    def _get_fg_color(self, color):
        fg_color = self._colors.get(color)
        if fg_color is None:
            return
        return '3{0}'.format(fg_color)

    def _get_bg_color(self, color):
        bg_color = self._colors.get(color)
        if bg_color is None:
            return
        return '4{0}'.format(bg_color)

    def colorize(self, text, fg=None, bg=None, attrs=None):
        _attrs = filter(None, (self._attrs.get(x) for x in (attrs or [])))
        _open_parts = list(_attrs)
        _open_parts.append(self._get_fg_color(fg))
        _open_parts.append(self._get_bg_color(bg))
        _open_parts = [x for x in _open_parts if x]

        if len(_open_parts) == 0:
            _open_parts.append(self._reset)
        t_open = '\033[{0}m'.format(';'.join(str(x) for x in _open_parts))
        t_close = '\033[{0}m'.format(self._reset)
        return u'{0}{1}{2}'.format(t_open, text, t_close)


def _256c(num):  # color
    return str(16 + int(num, 6))


def _256g(num):  # gray
    return str(231 + num)


class Xterm256Colorer(Xterm16Colorer):

    _colors = dict(
        grey=_256g(9),
        red=_256c('500'),
        green=_256c('130'),
        yellow=_256c('510'),
        blue=_256c('025'),
        magenta=_256c('515'),
        cyan=_256c('033'),
        white=_256g(21),
        hi_grey=_256g(14),
        hi_red=_256c('511'),
        hi_green=_256c('450'),
        hi_yellow=_256c('530'),
        hi_blue=_256c('135'),
        hi_magenta=_256c('503'),
        hi_cyan=_256c('244'),
        hi_white=_256g(24))

    def _get_fg_color(self, color):
        fg_color = self._colors.get(color)
        if fg_color is None:
            return
        return '38;5;{0}'.format(fg_color)

    def _get_bg_color(self, color):
        bg_color = self._colors.get(color)
        if bg_color is None:
            return
        return '48;5;{0}'.format(bg_color)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import datetime
import logging
import os
import sys
import six

from nicelog.colorers.terminal import get_term_colorer
from nicelog.styles.base import BaseStyle as DefaultStyle


POWERLINE_STYLE = bool(os.environ.get('HAS_POWERLINE_FONT', False))
DEFAULT = object()


class Colorful(logging.Formatter, object):

    def __init__(self, show_date=True, show_function=True,
                 show_filename=True, message_inline=False,
                 beautiful_tracebacks=True,
                 colorer=DEFAULT, style=DEFAULT, *a, **kw):

        """Log formatter for beautiful colored output

        Args
            show_date: whether to include date in log output
            show_function: whether to include function name
            show_filename: whether to include file name
            message_inline: whether to print messages inline
            beautiful_tracebacks: whether to nicely format tracebacks
            colorer (BaseColorer): used to create color output
            style (BaseStyle): color style to use
        """

        super(Colorful, self).__init__(*a, **kw)
        if style is DEFAULT:
            style = DefaultStyle()
        self.style = style
        if colorer is DEFAULT:
            colorer = get_term_colorer(style=style)
        self.colorer = colorer
        self._show_date = show_date
        self._show_function = show_function
        self._show_filename = show_filename
        self._message_inline = message_inline
        self._beautiful_tracebacks = beautiful_tracebacks

    def format(self, record):

        parts = []

        parts.append(self._format_level(record))

        if self._show_date:
            parts.append(self._format_date(record))

        parts.append(self._format_name(record))

        if self._show_filename:
            parts.append(self._format_filename(record))

        if self._show_function:
            parts.append(self._format_function(record))

        if self._message_inline:
            parts.append(self._format_message_inline(record).rstrip())
        else:
            parts.append(self._format_message_block(record).rstrip())

        # todo: beautiful exceptions if required to
        exc_info = self._format_traceback(record)
        if exc_info is not None:
            parts.append("\n\n" + self._indent(exc_info))

        return ' '.join(parts)

    def _render(self, style_name, text):
        if self.colorer is None:
            return text
        return self.colorer.render(style_name, text)

    def _format_date(self, record):
        fmtdate = datetime.datetime.fromtimestamp(
            record.created).strftime("%Y-%m-%d %H:%M:%S")
        return self._render('date', fmtdate)

    def _format_level_and_name(self, record):
        return ''.join((self._format_level(record), self._format_name(record)))

    def _format_level(self, record):
        return self._render(
            ('level_name_{0}'.format(record.levelname), 'level_name_DEFAULT'),
            ' {0:^8} '.format(record.levelname))

    def _format_name(self, record):
        return self._render('logger_name', ' {0} '.format(record.name))

    def _format_filename(self, record):
        return ':'.join((
            self._render('file_name', record.filename),
            self._render('line_number', str(record.lineno)),
        ))

    def _format_function(self, record):
        return '.'.join((
            self._render('module_name', record.module),
            self._render('function_name', str(record.funcName)),
        ))

    def _format_message_inline(self, record):
        return self._render(
            ('message_{0}'.format(record.levelname), 'message_DEFAULT'),
            record.getMessage().rstrip())

    def _format_message_block(self, record):
        return "\n" + self._indent(record.getMessage())

    def _format_traceback(self, record):
        if record.exc_info:
            if self._beautiful_tracebacks:
                return self._format_beautiful_traceback(record)
            text = self._format_plain_traceback(record)
            return self._render('exception', text)
        return None

    def _format_plain_traceback(self, record):
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            try:
                return six.u(record.exc_text)
            except UnicodeError:
                # Sometimes filenames have non-ASCII chars, which can lead
                # to errors when s is Unicode and record.exc_text is str
                # See issue 8924.
                # We also use replace for when there are multiple
                # encodings, e.g. UTF-8 for the filesystem and latin-1
                # for a script. See issue 13232.
                return record.exc_text.decode(
                    sys.getfilesystemencoding(), 'replace')
        return None

    def _format_beautiful_traceback(self, record):
        if not record.exc_info:
            return
        from nicelog.utils import TracebackInfo
        # todo: use colorer to render traceback!
        # todo: use suitable pygments formatter for the colorer
        trace = TracebackInfo.from_tb(record.exc_info[2]).format_color()
        trace += "\n" + repr(record.exc_info[1])
        return trace

    def _indent(self, text, tab="    ", level=1):
        _indent = tab * level
        lines = text.splitlines()
        indented = ["{0}{1}".format(_indent, line) for line in lines]
        return "\n".join(indented)


ColorLineFormatter = Colorful



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import io
import linecache
import sys

import six
from six.moves import zip, range


def trim_string(s, maxlen=1024, ellps='...'):
    """
    Trim a string to a maximum length, adding an "ellipsis"
    indicator if the string was trimmed
    """

    # todo: allow cutting in the middle of the string,
    #       instead of just on the right end..?

    if len(s) > maxlen:
        return s[:maxlen - len(ellps)] + ellps
    return s


class FrameInfo(object):
    def __init__(self, filename, lineno, name, line, locs):
        self.filename = filename
        self.lineno = lineno
        self.name = name
        self.line = line
        self.locs = self._format_locals(locs)
        self.context = self._get_context()

    def _get_context(self, size=3):
        """Return some "context" lines from a file"""
        _start = max(0, self.lineno - size - 1)
        _end = self.lineno + size
        _lines = linecache.getlines(self.filename)[_start:_end]
        _lines = [x.rstrip() for x in _lines]
        _lines = list(zip(range(_start + 1, _end + 1), _lines))
        return _lines

    def _format_locals(self, locs):
        formatted = {}
        for k, v in six.iteritems(locs):
            try:
                fmtval = trim_string(repr(v), maxlen=1024)
            except Exception as e:
                fmtval = '<Error creating repr(): {0}>'.format(repr(e))
            formatted[k] = fmtval
        return formatted


class TracebackInfo(object):
    """
    Class used to hold information about an error traceback.

    This is meant to be serialized & stored in the database, instead
    of a full traceback object, which is *not* serializable.

    It holds information about:

    - the exception that caused the thing to fail
    - the stack frames (with file / line number, function and exact code
      around the point in which the exception occurred)
    - a representation of the local variables for each frame.

    A textual representation of the traceback information may be
    retrieved by using ``str()`` or ``unicode()`` on the object
    instance.
    """

    def __init__(self):
        self.frames = []

    @classmethod
    def from_current_exc(cls):
        """
        Instantiate with traceback from ``sys.exc_info()``.
        """
        return cls.from_tb(sys.exc_info()[2])

    @classmethod
    def from_tb(cls, tb):
        """
        Instantiate from a traceback object.
        """
        obj = cls()
        obj.frames = cls._extract_tb(tb)
        return obj

    def format(self):
        """Format traceback for printing"""

        output = io.StringIO()
        output.write('------------------ '
                     'Traceback (most recent call last) '
                     '-----------------\n\n')
        output.write('\n'.join(
            self._format_frame(f)
            for f in self.frames))
        return output.getvalue()

    def format_color(self):  # todo: accept a colorizer + style
        """Format traceback for printing on 256-color terminal"""

        output = io.StringIO()
        output.write('\033[0m------------------ '
                     'Traceback (most recent call last) '
                     '-----------------\n\n')
        output.write(u'\n'.join(
            self._format_frame_color(f)
            for f in self.frames))
        return output.getvalue()

    def _format_frame(self, frame):
        output = io.StringIO()
        output.write(
            u'  File "{0}", line {1}, in {2}\n'.format(
                frame.filename, frame.lineno, frame.name))

        if frame.context:
            for line in frame.context:
                fmtstring = u'{0:4d}: {1}\n'
                if line[0] == frame.lineno:
                    fmtstring = u'    > ' + fmtstring
                else:
                    fmtstring = u'      ' + fmtstring
                output.write(fmtstring.format(line[0], line[1]))

        if len(frame.locs):
            output.write(u'\n      Local variables:\n')

            for key, val in sorted(six.iteritems(frame.locs)):
                output.write(u'        {0} = {1}\n'.format(key, val))

        return output.getvalue()

    def _format_frame_color(self, frame):
        from pygments import highlight
        from pygments.lexers import get_lexer_by_name
        from pygments.formatters import Terminal256Formatter

        _code_lexer = get_lexer_by_name('python')
        _code_formatter = Terminal256Formatter(style='monokai')

        def _highlight(code):
            return highlight(code, _code_lexer, _code_formatter)

        output = io.StringIO()
        output.write(
            u'\033[0m'
            u'\033[1mFile\033[0m \033[38;5;70m"{0}"\033[39m, '
            u'\033[1mline\033[0m \033[38;5;190m{1}\033[39m, '
            u'\033[1min\033[0m \033[38;5;214m{2}\033[0m\n\n'
            .format(frame.filename, frame.lineno, frame.name))

        if frame.context:
            for line in frame.context:
                fmtstring = u'{0:4d}: {1}\n'
                if line[0] == frame.lineno:
                    fmtstring = (u'    \033[48;5;250m\033[38;5;232m'
                                 u'{0:4d}\033[0m {1}\n')
                else:
                    fmtstring = (u'    \033[48;5;237m\033[38;5;250m'
                                 u'{0:4d}\033[0m {1}\n')

                color_line = _highlight(line[1])
                output.write(fmtstring.format(line[0], color_line.rstrip()))

        if len(frame.locs):
            output.write(u'\n    \033[1mLocal variables:\033[0m\n')

            for key, val in sorted(six.iteritems(frame.locs)):
                code_line = _highlight(u'{0} = {1}'.format(key, val)).rstrip()
                output.write(u'      {0}\n'.format(code_line))

        return output.getvalue()

    @classmethod
    def _extract_tb(cls, tb, limit=None):
        if limit is None:
            if hasattr(sys, 'tracebacklimit'):
                limit = sys.tracebacklimit
        frames = []
        n = 0
        while tb is not None and (limit is None or n < limit):
            f = tb.tb_frame
            lineno = tb.tb_lineno
            co = f.f_code
            filename = co.co_filename
            name = co.co_name
            linecache.checkcache(filename)
            line = linecache.getline(filename, lineno, f.f_globals)
            locs = f.f_locals  # Will be converted to repr() by FrameInfo
            if line:
                line = line.strip()
            else:
                line = None
            frames.append(FrameInfo(filename, lineno, name, line, locs))
            tb = tb.tb_next
            n = n+1
        return frames

    def __str__(self):
        return self.format().encode('utf-8')

    def __unicode__(self):
        return self.format()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

import six


class BaseColorer(object):
    def __init__(self, style):
        self.style = style

    def render(self, style, text):
        item_style = self._get_style(style)
        return self.colorize(text, **item_style)

    def _get_style(self, style_name):
        if isinstance(style_name, six.string_types):
            style_name = [style_name]
        for name in style_name:
            try:
                return getattr(self.style, name)
            except:
                pass
        return {}

    def colorize(self, text, fg=None, bg=None, attrs=None):
        return text



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, unicode_literals


class BaseStyle(object):

    date = dict(fg='cyan')

    level_name_DEBUG = dict(fg='cyan', attrs=['reverse'])
    level_name_INFO = dict(fg='green', attrs=['reverse'])
    level_name_WARNING = dict(fg='yellow', attrs=['reverse'])
    level_name_ERROR = dict(fg='red', attrs=['reverse'])
    level_name_CRITICAL = dict(fg='red', attrs=['reverse'])
    level_name_DEFAULT = dict(fg='white', attrs=['reverse'])

    logger_name = dict(fg='red', bg='white')
    file_name = dict(fg='green')
    line_number = dict(fg='hi_green')
    module_name = dict(fg='yellow')
    function_name = dict(fg='hi_yellow')

    message_DEBUG = dict(fg='cyan')
    message_INFO = dict(fg='green')
    message_WARNING = dict(fg='yellow')
    message_ERROR = dict(fg='red')
    message_CRITICAL = dict(fg='red')
    message_DEFAULT = dict(fg='white')

    exception = dict(fg='hi_grey')
